package pt.moth;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Iterator;

import soot.PackManager;
import soot.PhaseOptions;
import soot.Scene;
import soot.SootClass;
import soot.Transform;
import soot.options.Options;

public class Main {
	// Comando para correr o MOTH na consola:
	// Example 1: java -cp bin/classes/;lib/soot-trunk.jar pt.moth.Main -cp bin/test testes.Coord03.Main
	// Example 2: ./moth.sh -cp bin/test testes.Coord03.Main
	
	public static final String DEFAULT_JAVA_PATH = System
			.getProperty("sun.boot.class.path")
			+ File.pathSeparator
			+ System.getProperty("java.class.path");
	public static final String DEFAULT_RES_PATH = "res" + File.separator;
	private static final String help_message = "Usage:\njava -cp bin/classes:lib/soot-trunk.jar pt.moth.Main [-options]\n./moth.sh\n\n"
			+ "Where options include:\n"
			+ "\t-r \t<path of XML resources>\n"
			+ "\t-cp \t<application class path> (MANDATORY)\n"
			+ "\t-? -h \tprints this help message\n"
			+ "\t-j \t<path to java run time classes ('rt.jar')>\n";

	private static void help() {
		System.out.println(help_message);
		System.exit(0);
	}

	private static void exitProgram() {
		System.err.println("Invalid Arguments!");
		System.err.println(help_message);
		System.exit(-1);
	}

	public static void main(String[] args) throws FileNotFoundException {
		if (args.length == 0) {
			exitProgram();
		}

		boolean r_defined = false; // ja adicionamos a "res"?
		boolean j_defined = false; // ja adicionamos o java path?
		boolean cp_defined = false; // ja adicionamos o class path?

		String main_class_name = args[args.length - 1];
		String java_path = "none";
		String cp_path = "none";
		String res_path = "none";

		for (int i = 0; i < args.length - 1; i++) {
			String arg = args[i];
			if (arg.equalsIgnoreCase("-r")) {
				i++; // args[i] has the res path
				if (i >= args.length || r_defined) {
					exitProgram();
				}
				res_path = args[i];
				r_defined = true;
			} else if (arg.equalsIgnoreCase("-cp")) {
				i++; // args[i] has the class path
				if (i >= args.length || cp_defined) {
					exitProgram();
				}
				cp_path = args[i];
				cp_defined = true;
			} else if (arg.equalsIgnoreCase("-j")) {
				i++; // args[i] has the java path
				if (i >= args.length || j_defined) {
					exitProgram();
				}
				java_path = args[i];
				j_defined = true;
			} else if (arg.equalsIgnoreCase("-?") || arg.equalsIgnoreCase("-h")) {
				help();
			} else {
				exitProgram();
			}
		}
		if (!cp_defined) {
			exitProgram();
		}

		// Run MOTH

		System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
		System.out.println("@@@@@       Main Test       @@@@@");
		System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
		System.out.println("(With class " + main_class_name + ")...\n\n");
		long init = System.currentTimeMillis();
		run_moth(main_class_name /* , soot_path */, java_path, cp_path, res_path);
		long end = System.currentTimeMillis();
		System.out.println("\n\n@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
		System.out.println("@@@         " + (end - init) / 1000
				+ "   seconds     @@@");
		System.out.println("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@");
		System.exit(0);
	}

	private static void run_moth(String main_class_name, String java_path,
			String cp_path, String res_path) {
		System.out.println("Starting...");

		/**
		 * Soot command-line options See: {@link http
		 * ://www.sable.mcgill.ca/soot/tutorial/usage/}
		 */

		// Activates whole program analysis (used in inter-procedural analysis)
		Options.v().set_whole_program(true);

		Options.v().set_allow_phantom_refs(true);

		// Excludes the analysis of classes under java and javax packages
		// Options.v().set_exclude(Arrays.asList(new String[] {"java",
		// "javax"}));

		Options.v().set_no_bodies_for_excluded(false);

		Options.v().set_verbose(false);

		Options.v().set_debug(false);

		Options.v().set_debug_resolver(false);

		// This option forces the dump of the CFG of each analyzed method
		// Options.v().set_dump_cfg(Arrays.asList(new String[] {"ALL"}));

		Options.v().set_keep_line_number(true);

		/**
		 * Internal Soot options
		 */
		String r_path = DEFAULT_RES_PATH;
		String j_path = DEFAULT_JAVA_PATH;

		if (!res_path.equals("none")) {
			r_path = res_path;
		}
		if (!java_path.equals("none")) {
			j_path = res_path;
		}

		Scene.v().setSootClassPath(j_path + File.pathSeparator + cp_path);

		PhaseOptions.v().setPhaseOption("jb", "use-original-names:true");
		PhaseOptions.v().setPhaseOption("jb.ulp", "enabled:false");

		PackManager
				.v()
				.getPack("wjtp")
				.add(new Transform("wjtp.myanalysis", new AnalysisTransformer(
						r_path + "/methods.xml")));

		Scene.v().addBasicClass(/* class_name */main_class_name,
				SootClass.SIGNATURES);
		Scene.v().loadNecessaryClasses();

		Iterator<SootClass> it = Scene.v().getClasses().iterator();
		while (it.hasNext()) {
			SootClass c = it.next();
			if (!(c.getJavaPackageName().startsWith("java.")
					|| c.getJavaPackageName().startsWith("javax.") || c
					.getJavaPackageName().startsWith("sun."))) {

				if (c.isPhantom()) {
					System.out.println("Class: " + main_class_name
							+ " not found!");
					System.exit(0);
				}
				c.setApplicationClass();
			}
		}
		SootClass c = Scene.v().loadClassAndSupport(main_class_name);

		// Needed when doing whole program analysis
		Scene.v().setMainClass(c);

		PackManager.v().runPacks();
	}
}
