package pt.moth.arthoApproach;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Map.Entry;


import pt.moth.processAnalysis.Process;
import pt.moth.util.View;
import pt.moth.util.ViewCardinalityComparator;
import pt.moth.util.conflicts.Conflict;
import pt.moth.util.conflicts.ConflictType;
import pt.moth.util.conflicts.ViewConflict;


/**
 * This class represents the view consistency algorithm of Artho
 * It has the same structure of our sensors
 * @author Vasco Pessanha
 *
 */
public class ArthoSensor{

	//For each Process, this atribute keeps the views of that process
	private Map<String, List<View>> views;
	//For each Process, this atribute keeps the maximalViews of that process
 	private Map<String, List<View>> maxViews;
	//List of all the processes
	private Set<Process> processes;
 
	public ArthoSensor(){
		views = new HashMap<String, List<View>>();
		maxViews = new HashMap<String, List<View>>();
		processes = new HashSet<Process>();
	}

	public void pushViewInformation() {
 		//For each process...
		for(Entry<String, List<View>> e : ArthoViewsCollector.getViews()){
			views.put(e.getKey(), e.getValue());
			computeMaximalViews(e.getKey());
		}
		processes = ArthoViewsCollector.getProcesses();
	}
 	/**
 	 * Computes a process' maximal views..
 	 * @param p process
 	 */
	private void computeMaximalViews(String p){
		boolean isMaximalView;
		List<View> allviews = views.get(p);
		List<View> maximalViews = new ArrayList<View>();
		for (View v : allviews) {
			isMaximalView = true;
			for (View v2 : allviews) {
				if (!v2.equalsIgnoreType(v) && v2.containsIgnoreType(v)) {
					isMaximalView = false;
					break;
				}
			}
			if (isMaximalView) {
				maximalViews.add(v);
			}
		}
		maxViews.put(p, maximalViews);
	}

	
	public Set<Conflict> getConflicts() {
		boolean allOfTheSameSize = 
			   (maxViews.size() == views.size())
			&& (maxViews.size() == processes.size());
		Set<Conflict> conflicts = new HashSet<Conflict>();
		if(!allOfTheSameSize){
			System.err.println("Error! (ArthoSensor) Every of this atributes should have the same size (equals to the number of the processes)");
		}
		else{
			Process[] array = new Process[processes.size()]; //could be with another atribute
			int pos = 0;
			for(Process p : processes){
				array[pos] = p;
				pos ++;
			}
			for(int i = 0; i < array.length; i++){
				for(int j = i; j < array.length; j++){
					getConflicts(conflicts, array[i], array[j]);
					if(i != j){
						getConflicts(conflicts, array[j], array[i]);
					}
				}
			}
		}
		return conflicts;
	}
	
	/**
	 * Compute all conflicts between p1's maximal views and p2's views
	 * @param conflicts is a reference to the set of all conflicts (return) 
	 * @param p1 first process (maximal views)
	 * @param p2 second process (views)
	 */
	private void getConflicts(Set<Conflict> conflicts, Process p1, Process p2){
		Set<View> overlappingViews;
		List<View> p1_maxViews = maxViews.get(p1.getName());
		List<View> p2_Views = views.get(p2.getName());
		View intersection = null;
		
		if (p1_maxViews != null && p2_Views != null) {
			for (View max_v : p1_maxViews) {
				overlappingViews = new HashSet<View>();
				for (View v : p2_Views) {
					intersection = v.intersection(max_v);
					if(intersection != null && intersection.size()>0 && 
							!intersection.getMethod().equals(max_v.getMethod())){
						overlappingViews.add(intersection.clone());
					}
				}
				if (!formChain(overlappingViews)) {
					conflicts.add(new ViewConflict(p1, p2, overlappingViews, max_v, ConflictType.ARTHO));
				}
			}
		}
	}
	/**
	 * Returns if the given set forms a chain.. For All members mi, the forms a chain if m0 C m1 C m2 C .. C mn
	 * 
	 * Algorithm:
	 * 		1) Order the list by view's size. If 2 views have the same size it's not important which is the bigger
	 * 		2) Iterate the list comparing in groups of 2
	 * 			2.1) If v2.size < v1.size		=>		ERROR! Should be ordered
	 * 			2.2) If v2.size >= v1.size		=>		v2.contains(v1)? (Yes => keep going)(No => Returns false)
	 * 		3) Return true at the end 
	 * 
	 * Since we work with Set's, if the view's sizes are equal than we can return false (they won't be equal)
	 *
	 * @param subsetViews - the given Set
	 * @return true if the set forms a chain and false otherwise 
	 */
	private boolean formChain(Set<View> subsetViews) {
		if(subsetViews.size()<= 1){
			return true;
		}
		TreeSet<View> orderedSet = new TreeSet<View>(new ViewCardinalityComparator());
		for(View v : subsetViews){
			orderedSet.add(v);
		}
		Object[] array = orderedSet.toArray();
		View v1,v2;
		for(int i = 0; i< orderedSet.size()-1;i++){
			v1 = (View)array[i];
			v2 = (View)array[i+1];
			if(v2.size()<v1.size()){
				System.err.println("ERROR! (Results Handler) - The Set should be ordered...");
				System.exit(-1);
			}
			if(!v2.containsIgnoreType(v1)){
				return false;
			}
		}
		return true;
	}
}
