package pt.moth.dependencies;

import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import pt.moth.processAnalysis.Process;
import pt.moth.processAnalysis.ProcessGenerator;
import pt.moth.util.Access;
import pt.moth.util.IAccess;
import pt.moth.util.View;
import pt.moth.util.variables.GlobalVar;
import pt.moth.viewsAnalysis.ViewsCollector;


/**
 * This class represents a Dependency between two views/methods/transactions. A
 * dependency is a triple (V1,var,V2) that correspond to a dependence between
 * views V1 and V2 with variable var. This occurs when we have the following
 * code: var = res(V1); if(...var...){ ...V2... }
 * 
 * A dependency can be either a dataDependency or a ControlDependency (if)
 * @author Vasco Pessanha
 * 
 */

public class Dep{
	//v1 --> v2
	protected View v1;
	protected View v2;
	protected GlobalVar var;
	protected Process process;
	protected boolean isControlDep;
	
	public Dep(View v1, GlobalVar var, View v2, boolean isControlDep){
		this.v1 = v1;
		this.var = var;
		this.v2 = v2;
		this.process = tryToGuessProcess().clone();
		this.isControlDep = isControlDep;
	}
	
	private Process tryToGuessProcess() {
		Set<Entry<String, List<View>>> r_views = ViewsCollector.getReadViews();
		Set<Entry<String, List<View>>> w_views = ViewsCollector.getWriteViews();
		List<String> processes = new ArrayList<String>(); 
		List<String> possibleReturns = new ArrayList<String>(); 
		for(Entry<String, List<View>> entry : r_views){
			List<View> list_views = entry.getValue();
			if(list_views.contains(trimWrites(v1))){
				processes.add(entry.getKey());
			}
		}
		for(Entry<String, List<View>> entry : w_views){
			if(processes.contains(entry.getKey())){
				List<View> list_views = entry.getValue();
				if(list_views.contains(trimReads(v2))){
					possibleReturns.add(entry.getKey());
				}
			}
		}
		if(possibleReturns.size() == 1)
			return ProcessGenerator.getProcess(possibleReturns.get(0));
		return new Process("UnKnown Process",
			"This process was generated in Dep's Class (code line = 71)");
	}

	/**
	 * This function trims all the write access of view
	 */
	private View trimWrites(View view) {
		View result = view.clone();
		List<IAccess> list = result.getAccesses().toList();
		for(IAccess a : list){
			if(a instanceof Access && ((Access)a).isWrite()){
				result.getAccesses().remove(a);
			}
		}
		return result;
	}

	/**
	 * This function trims all the write access of view
	 */
	private View trimReads(View view) {
		View result = view.clone();
		List<IAccess> list = result.getAccesses().toList();
		for(IAccess a : list){
			if(a instanceof Access && ((Access)a).isRead()){
				result.getAccesses().remove(a);
			}
		}
		return result;
	}
	public String toString(){
		return v1.getMethod().getName()+" -> " +v2.getMethod().getName()+"("+var.getName()+")";
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((v1 == null) ? 0 : v1.hashCode());
		result = prime * result + ((v2 == null) ? 0 : v2.hashCode());
		result = prime * result + ((var == null) ? 0 : var.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Dep other = (Dep) obj;
		if (v1 == null) {
			if (other.v1 != null)
				return false;
		} else if (!v1.equals(other.v1))
			return false;
		if (v2 == null) {
			if (other.v2 != null)
				return false;
		} else if (!v2.equals(other.v2))
			return false;
		if (var == null) {
			if (other.var != null)
				return false;
		} else if (!var.equals(other.var))
			return false;
		return true;
	}

	public View getFirstView() {
		return v1;
	}

	public View getSecondView() {
		return v2;
	}

	public GlobalVar getDepVariable() {
		return var;
	}

	public Process getProcess() {
		return process;
	}
}
