package pt.moth.dependencies;


import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import pt.moth.inter.MethodUtils;
import pt.moth.util.Pair;
import pt.moth.util.graphs.DepEdge;
import pt.moth.util.graphs.DirectedGraph;
import pt.moth.util.variables.GlobalVar;
import pt.moth.util.variables.IGlobalVar;
import pt.moth.util.variables.NativeGlobalVar;
import pt.moth.util.variables.ParamRef;
import pt.moth.util.variables.ParamVar;
import pt.moth.util.variables.ReturnRef;
import pt.moth.util.variables.ReturnVar;
import pt.moth.util.variables.Variable;

import soot.SootMethod;

public class VariableGraph extends DirectedGraph<VarVersion, DepEdge> {

	/**
	 * Return a list of all important vertices of the graph. A vertice is an
	 * important vertex iff it is a global variable of an atomic method, or a
	 * parameter variable of an atomic method
	 * 
	 * This list is returned as a map, since it is faster to get an element
	 */
	public Map<VarVersion, Boolean> getImportantVertices() {
		Map<VarVersion, Boolean> result = new HashMap<VarVersion, Boolean>();
		for (VarVersion node : vertices) {
			Variable var = node.getVariable();
			boolean isImportant = false;
			if (var instanceof NativeGlobalVar) {
				isImportant = true;
			} else if (var instanceof GlobalVar
					&& MethodUtils.methodIsAtomic(((Variable) var)
							.getSootMethod())) {
				isImportant = true;
			}
			else if(var instanceof ReturnRef
					&& MethodUtils.methodIsAtomic(((Variable) var)
							.getSootMethod())) {
				isImportant = true;
			}
			if (isImportant) {
				result.put(node, true);
			}
		}
		return result;
	}

	/**
	 * Returns a list of all vertices that are in "set", and are reachable from
	 * "origin"
	 * 
	 * @param origin
	 *            - Vertice from which we reach the returned vertices
	 * @param set
	 *            - Set of vertices (could be a list or set but it is faster
	 *            with a map structure)
	 * @return subset of "set" which contain all vertices of "set" that are
	 *         reachable from origin
	 */
	public List<Pair<VarVersion, Boolean>> getImportantReached(VarVersion origin,
			Map<VarVersion, Boolean> set) {
		List<Node> treated = new ArrayList<Node>();
		List<Node> toExplore = new ArrayList<Node>();

		List<Pair<VarVersion, Boolean>> result = new ArrayList<Pair<VarVersion, Boolean>>();
		String toDebug = "aaaaaa"; // The path with this var as origin will be printed

		// When we reach to a ReturnVar we connect it with the respective
		// ReturnRef
		Map<SootMethod, ReturnRef> returnRefs = new HashMap<SootMethod, ReturnRef>();

		SootMethod originMethod = origin.getSootMethod();
		// Put the origin's sucessors
		toExplore.add(new Node(origin));

		while (toExplore.size() > 0) {
			Node node = toExplore.remove(0).clone();

			if (!treated.contains(node)) {
				treated.add(node);
				if (shouldAdd(node, originMethod, set)) {
					if (origin.toString().contains(toDebug)) {
						System.out.println("ADD! " + node);
					}
					result.add(new Pair<VarVersion, Boolean>(node.getGraphNode(),
							node.hasControlDep()));
				}

				if (origin.toString().contains(toDebug)) {
					System.out.println("\n" + node);
				}
				Variable v = node.getVariable();
				Set<DepEdge> succs = getSuccEdges(node.getGraphNode());
				if (succs.size() == 0) {
					if (v instanceof ParamRef) {
						ParamRef param = (ParamRef) v;

						// Store the ReturnReference so we can come back of the
						// method call
						ReturnRef ret = new ReturnRef(param.getCallerMethod(),
								param.getInvStmt());
						returnRefs.put(param.getSootMethod(), ret);

						// Jump to the real ParamVariable
						Node succ = node.clone();
						succ.setGraphNode(new VarVersion(param.getParamVar()));
						toExplore.add(0, succ);
						if (origin.toString().contains(toDebug)) {
							System.out.println("   " + succ);
						}
					} else if (v instanceof ReturnVar) {
						ReturnRef ref = returnRefs.get(v.getSootMethod());
						if (ref != null) {
							Node succ = node.clone();
							succ.setGraphNode(new VarVersion(ref));
							toExplore.add(0, succ);
							if (origin.toString().contains(toDebug)) {
								System.out.println("   REF = " + succ);
							}
						} else {
							Set<ReturnRef> allRefs = getAllReturnRefs(v
									.getSootMethod());
							Node succ;
							for (ReturnRef r : allRefs) {
								succ = node.clone();
								succ.setGraphNode(new VarVersion(r));
								toExplore.add(0, succ);
								if (origin.toString().contains(toDebug)) {
									System.out.println("   ALLREF = " + succ);
								}
							}
						}
					}
				}
				for (DepEdge e : succs) {
					Node succ = node.getNext(e, originMethod);
					if (!treated.contains(succ)) {
						if (origin.toString().contains(toDebug)) {
							System.out.println("   " + succ);
						}
						toExplore.add(0, succ);
					}
				}
			}
		}
		return result;
	}

	/**
	 * Should add (returns true) iff: 
	 * - is a ControlDependency (integer == 3) and has a different method
	 * - is a DataDependency but passed through the
	 * returnVar and the paramVar of the method (integer == 2)
	 * 
	 * @param set
	 */
	private boolean shouldAdd(Node node, SootMethod originMethod,
			Map<VarVersion, Boolean> set) {
		VarVersion gNode = node.getGraphNode();
		Variable v = gNode.getVariable();
		if ((v instanceof ReturnRef || v instanceof IGlobalVar) && set.containsKey(gNode)) {
			if (node.hasControlDep()
					&& !originMethod.equals(node.getVariable().getSootMethod())) {
				return true;
			}
			if (node.passedReturnVar()) {
				SootMethod m = v instanceof NativeGlobalVar ? ((NativeGlobalVar) v)
						.getAtomicMethod() : v.getSootMethod();
				Set<VarVersion> params = getMethodParamVars(m);
				for (VarVersion p : params) {
					if (hasPath(p, gNode)) {
						return true;
					}
				}
			}
		}
		return false;
	}

	private Set<VarVersion> getMethodParamVars(SootMethod m) {
		Set<VarVersion> result = new HashSet<VarVersion>();
		for (VarVersion node : vertices) {
			Variable v = node.getVariable();
			if (v instanceof ParamVar && v.getSootMethod().equals(m)) {
				result.add(node.copy());
			}
		}
		return result;
	}

//	/**
//	 * Everytime we have a local vertice we connect its predecessors its
//	 * sucessors. Then we erase every edge with that vertice
//	 */
//	public void pruneLocalVertices() {
//		Variable local = findFirstLocalVertice();
//		while (local != null) {
//
//			// link up all predecessor with all successor
//			linkUp(getPredEdges(local), getSuccEdges(local));
//
//			// Remove the local vertice and its edges
//			removeEdgeWithVertex(local);
//			removeVertex(local);
//
//			// Keep going with the next local edge
//			local = findFirstLocalVertice();
//		}
//	}

//	/**
//	 * Link up every vertice from "preds" to every vertice of "succs"
//	 * 
//	 * @param preds
//	 *            - predecessors of the first vertice of a given edge
//	 * @param succs
//	 *            - successors of the second vertice of a given edge
//	 */
//	private void linkUp(Set<DepEdge> preds, Set<DepEdge> succs) {
//		for (DepEdge p : preds) {
//			for (DepEdge s : succs) {
//				addEdge(new DepEdge(p.getOrigin(), s.getDest(),
//						p.isControlDep() || s.isControlDep()));
//			}
//		}
//	}
//
//	private GraphNode findFirstLocalVertice() {
//		for (GraphNode e : vertices) {
//			if (e.getVariable() instanceof LocalVar) {
//				return e;
//			}
//		}
//		return null;
//	}

	/**
	 * Return every ReturnRef Variable of "m". In other words, for every time
	 * that "m" is called, this method returns a reference of that reference
	 */
	public Set<ReturnRef> getAllReturnRefs(SootMethod m) {
		Set<ReturnRef> result = new HashSet<ReturnRef>();
		for (VarVersion node : vertices) {
			Variable v = node.getVariable();
			if (v instanceof ReturnRef
					&& (((ReturnRef) v).getSootMethod()).equals(m)) {
				result.add((ReturnRef) v);
			}
		}
		return result;
	}

//	/**
//	 * Prune every ParamVar from this graph
//	 */
//	public void pruneParams() {
//		for (GraphNode node : vertices) {
//			Variable v = node.getVariable();
//			if (v instanceof ParamVar) {
//				Set<DepEdge> succs = getSuccEdges(node);
//				Set<DepEdge> preds = getPredEdges(node);
//				for (DepEdge p : preds) {
//					for (DepEdge s : succs) {
//						addEdge(new DepEdge(p.getOrigin(), s.getDest(),
//								p.isControlDep() || s.isControlDep()));
//					}
//				}
//				removeEdgeWithVertex(v);
//			}
//		}
//	}

	/**
	 * We have to rewrite this method, so we can jump from "ParamRef" to the
	 * parameters and from the ReturnVar to the ReturnRefs
	 */
	@Override
	public boolean hasPath(VarVersion origin, VarVersion dest){
		List<Node> treated = new ArrayList<Node>();
		List<Node> toExplore = new ArrayList<Node>();
		
		// When we reach to a ReturnVar we connect it with the respective
		// ReturnRef
		Map<SootMethod, ReturnRef> returnRefs = new HashMap<SootMethod, ReturnRef>();

		SootMethod originMethod = origin.getVariable().getSootMethod();
		// Put the origin's sucessors
		toExplore.add(new Node(origin));

		while (toExplore.size() > 0) {
			Node node = toExplore.remove(0).clone();

			if (!treated.contains(node)) {
				treated.add(node);
				Variable v = node.getVariable();

				if (v.equals(dest)) {
					return true;
				}

				Set<DepEdge> succs = getSuccEdges(node.getGraphNode());
				if (succs.size() == 0) {
					if (v instanceof ParamRef) {
						ParamRef param = (ParamRef) v;

						// Store the ReturnReference so we can come back of the
						// method call
						ReturnRef ret = new ReturnRef(param.getCallerMethod(),
								param.getInvStmt());
						returnRefs.put(param.getSootMethod(), ret);

						// Jump to the real ParamVariable
						Node succ = node.clone();
						succ.setGraphNode(new VarVersion(param.getParamVar()));
						toExplore.add(0, succ);
					} else if (v instanceof ReturnVar) {
						ReturnRef ref = returnRefs.get(v.getSootMethod());
						if (ref != null) {
							Node succ = node.clone();
							succ.setGraphNode(new VarVersion(ref));
							toExplore.add(0, succ);
						} else {
							Set<ReturnRef> allRefs = getAllReturnRefs(v
									.getSootMethod());
							Node succ;
							for (ReturnRef r : allRefs) {
								succ = node.clone();
								succ.setGraphNode(new VarVersion(r));
								toExplore.add(0, succ);
							}
						}
					}
				}
				for (DepEdge e : succs) {
					Node succ = node.getNext(e, originMethod);
					if(succ.getGraphNode().equals(dest)){
						return true;
					}
					if (!treated.contains(succ)) {
						toExplore.add(0, succ);
					}
				}
			}
		}
		return false;
	}

	/**
	 * Returns every variable that can be reached from the variable "origin"
	 */
	public Set<VarVersion> getReachedVars(VarVersion origin){
		Set<VarVersion> initialSet = new HashSet<VarVersion>();
		initialSet.add(origin);
		return getReachedVars(initialSet);
	}
	/**
	 * Returns every variable that is reached from one of the variables of the "initialSet"
	 */
	public Set<VarVersion> getReachedVars(Set<VarVersion> initialSet){
		List<VarVersion> treated = new ArrayList<VarVersion>();
		List<VarVersion> toExplore = new ArrayList<VarVersion>();
		Set<VarVersion> result = new HashSet<VarVersion>();
		
		// When we reach to a ReturnVar we connect it with the respective
		// ReturnRef
		Map<SootMethod, ReturnRef> returnRefs = new HashMap<SootMethod, ReturnRef>();

		// Put the origin's sucessors
		for(VarVersion v : initialSet)
			toExplore.add(v);

		while (toExplore.size() > 0) {
			VarVersion v = toExplore.remove(0).copy();
			result.add(v);
			
			if (!treated.contains(v)) {
				treated.add(v);
				
				Set<DepEdge> succs = getSuccEdges(v);
				if (succs.size() == 0) {
					Variable var = v.getVariable();
					if (var instanceof ParamRef) {
						ParamRef param = (ParamRef) var;

						// Store the ReturnReference so we can come back of the
						// method call
						ReturnRef ret = new ReturnRef(param.getCallerMethod(),
								param.getInvStmt());
						returnRefs.put(param.getSootMethod(), ret);

						// Jump to the real ParamVariable
						toExplore.add(0, new VarVersion(param.getParamVar()));
					} else if (var instanceof ReturnVar) {
						ReturnRef ref = returnRefs.get(var.getSootMethod());
						if (ref != null) {
							toExplore.add(new VarVersion(ref));
						} else {
							Set<ReturnRef> allRefs = getAllReturnRefs(var.getSootMethod());
							for (ReturnRef r : allRefs) {
								toExplore.add(new VarVersion(r));
							}
						}
					}
				}
				for (DepEdge e : succs) {
					VarVersion succ = e.getDest();
					if (!treated.contains(succ)) {
						toExplore.add(succ);
					}
				}
			}
		}
		return result;
	}
}

/**
 * This class represents a Node of the paths used in the getImportantReached
 * method.
 */
class Node {
	// Last Variable passed in this path
	private VarVersion v;
	// If it already passed the ReturnVar of the originMethod
	private boolean passedReturn;
	// If it already passed an edge that represented a control dependency
	private boolean hasControlDep;

	private Node(VarVersion v2, boolean passedReturn, boolean hasControlDep) {
		this(v2);
		this.passedReturn = passedReturn;
		this.hasControlDep = hasControlDep;
	}

	public boolean passedReturnVar() {
		return passedReturn;
	}

	/**
	 * Returns the next Node of the path, having in count if it now passed a
	 * ParamVar, ReturnVar, ControlDep, etc..
	 */
	public Node getNext(DepEdge e, SootMethod originMethod) {
		// e = (v, resultNode, ?)
		VarVersion node = e.getDest();
		Node result = new Node(node, false, false);
		Variable v = node.getVariable();
		result.passedReturn = passedReturn
				|| ((v instanceof ReturnVar || v instanceof ReturnRef) && v
						.getSootMethod().equals(originMethod));

		result.hasControlDep = hasControlDep
				|| (e.isControlDep() && passedReturn);

		return result;
	}

	public void setGraphNode(VarVersion newVar) {
		v = newVar.copy();
	}

	public Node(VarVersion v) {
		this.v = v.copy();
		passedReturn = false;
		hasControlDep = false;
	}

	public VarVersion getGraphNode() {
		return v;
	}

	public Variable getVariable() {
		return v.getVariable();
	}
	
	public boolean hasControlDep() {
		return hasControlDep;
	}

	public Node clone() {
		return new Node(v, passedReturn, hasControlDep);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + (hasControlDep ? 1231 : 1237);
		result = prime * result + (passedReturn ? 1231 : 1237);
		result = prime * result + ((v == null) ? 0 : v.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Node other = (Node) obj;
		if (hasControlDep != other.hasControlDep)
			return false;
		if (passedReturn != other.passedReturn)
			return false;
		if (v == null) {
			if (other.v != null)
				return false;
		} else if (!v.equals(other.v))
			return false;
		return true;
	}

	@Override
	public String toString() {
		return "Node(" + v + "," + passedReturn + "," + hasControlDep + ")";
	}
}
