package pt.moth.dependencies.secondPhase;

import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Stack;

import pt.moth.util.variables.IfConditionVar;
import pt.moth.util.variables.ReturnRef;

import soot.toolkits.scalar.AbstractFlowSet;
import soot.toolkits.scalar.FlowSet;

/**
 * This FlowSet/Lattice keeps a set of dependency trees.
 * This allows us to represent all data dependencies of a program
 * @author Vasco Pessanha
 * 
 */
public class StackLattice extends AbstractFlowSet {

	protected Stack<SetIfVars> ifVars;
	protected SetIfVars old_ifVars;
	protected Set<ReturnRef> refs;
		
	public StackLattice(SetIfVars set, Set<ReturnRef> refs){
		this.ifVars = new Stack<SetIfVars>();
		this.old_ifVars = set.copy();
		this.refs = refs == null? new HashSet<ReturnRef>() : refs;
	}
	public StackLattice(Set<ReturnRef> refs){
		this.ifVars = new Stack<SetIfVars>();
		this.old_ifVars = new SetIfVars();
		this.refs = refs == null? new HashSet<ReturnRef>() : refs;
	}
	
	public Set<ReturnRef> getReturnRefs() {
		return refs;
	}
	public void addReturnRef(ReturnRef returnRef) {
		if(returnRef != null)
			this.refs.add(returnRef);
	}
	
	public void add(Object param) {
		if(param instanceof SetIfVars){
			SetIfVars set = (SetIfVars)param;
			if(!ifVars.contains(set)){
				ifVars.add((SetIfVars)param);
			}
		}
		else{
			System.err.println("(StackLattice) ERROR! - Shouldn't happen PARAM = "+param);
		}
	}
	public StackLattice clone() {
		StackLattice lat = new StackLattice(old_ifVars,refs);
		for(SetIfVars s : ifVars){
			lat.add(s);
		}
		return lat;
	}
	
	public boolean contains(Object param) {
		if(param instanceof SetIfVars){
			return ifVars.contains((SetIfVars)param);
		}
		return false;
	}

	public boolean isEmpty() {
		return ifVars.isEmpty() && old_ifVars.isEmpty();
	}

	public int size() {
		return ifVars.size() + old_ifVars.size();
	}
	public List<SetIfVars> toList() {
		return ifVars;
	}

	@Override
	public void clear() {
		ifVars.clear();
		old_ifVars.clear();
		refs = null;
	}

	@Override
	public void copy(FlowSet dest) {
		StackLattice out = (StackLattice)dest;
		out.clear();
		out.refs = this.refs;
		out.setOldVars(old_ifVars);
		for(SetIfVars s : ifVars){
			out.add(s.copy());	
		}
	}

	private void setOldVars(SetIfVars set) {
		old_ifVars = set.copy();
	}
	public void union(FlowSet set){
		StackLattice lat = (StackLattice) set;
		
		for(ReturnRef r : lat.refs){
			addReturnRef(r);
		}
		for(SetIfVars s : lat.ifVars){
			add(s);
		}
		for(IfConditionVar var : lat.old_ifVars){
			this.old_ifVars.add(var);
		}
	}
	@Override
	public void union(FlowSet set1, FlowSet dest) {
		StackLattice out = (StackLattice)dest;
		
		out.clear();
		out.refs = this.refs;
		for(ReturnRef r : ((StackLattice)set1).refs){
			out.addReturnRef(r);
		}
		
		out.setOldVars(old_ifVars);
		for(IfConditionVar var : ((StackLattice)set1).old_ifVars){
			out.old_ifVars.add(var);
		}
		for(SetIfVars s : ifVars){
			out.add(s);
		}
		for(SetIfVars s : ((StackLattice)set1).ifVars){
			out.add(s);
		}
	}
	
	@Override
	public Object emptySet() {
		return new StackLattice(null);
	}

	public String toString(){
		return "{Vars = "+varsToString()+"\nOlds = "+oldsToString()+"}\nReturnRefs = {"+refs+"}";
	}
	
	private String varsToString(){
		String r = "{";
		if(ifVars.size() == 0)
			return r+"}";
		for(SetIfVars s : ifVars){
			r+=s.toString()+",";
		}
		return r.substring(0,r.length()-1)+"}";
	}
	private String oldsToString(){
		String r = "{";
		if(old_ifVars.size() == 0)
			return r+"}";
		for(IfConditionVar var : old_ifVars){
			r+=var.toString()+",";
		}
		return r.substring(0,r.length()-1)+"}";
	}
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result
				+ ((refs == null) ? 0 : refs.hashCode());
		result = prime * result
				+ ((ifVars == null) ? 0 : ifVars.hashCode());
		result = prime * result
			+ ((old_ifVars == null) ? 0 : old_ifVars.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if(!(obj instanceof StackLattice)){
			return false;
		}
		StackLattice other = (StackLattice)obj;
		if(!compareStacks(ifVars,other.ifVars))
			return false;
		if(old_ifVars == null){
			if(other.old_ifVars != null){
				return false;
			}
		}
		if(!old_ifVars.equals(other.old_ifVars))
			return false;
		if(refs == null)
			return other.refs == null;
		if(!refs.equals(other.refs))
			return false;
		return true;
	}

	@SuppressWarnings("unchecked")
	private boolean compareStacks(Stack<SetIfVars> st1, Stack<SetIfVars> st2) {
		if(st1.size() != st2.size())
			return false;
		int size = st1.size();
		Stack<SetIfVars> cl1 = (Stack<SetIfVars>) st1.clone();
		Stack<SetIfVars> cl2 = (Stack<SetIfVars>) st2.clone();
		for(int i = 0;i<size; i++){
			if(!cl1.pop().equals(cl2.pop())){
				return false;
			}
		}
		return true;
	}
	
	@Override
	public Iterator<SetIfVars> iterator() {
		return ifVars.iterator();
	}
	
	public SetIfVars removeLast() {
		if(!ifVars.empty()){
			return ifVars.pop();
		}
		return null;
	}

	@Override
	public void intersection(FlowSet in1, FlowSet out) {
		SetIfVars set = ((StackLattice)in1).old_ifVars;
		out.clear();
		((StackLattice)out).setOldVars(set);
		for(SetIfVars s : ifVars){
			if(in1.contains(s)){
				out.add(s);
			}
		}
		for(IfConditionVar var : old_ifVars){
			((StackLattice)out).old_ifVars.add(var);
		}
	}
	@Override
	public void add(Object paramObject, FlowSet paramFlowSet) {
		System.err.println("(StackLattice)ERROR! - method add(Object paramObject, FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void difference(FlowSet paramFlowSet) {
		System.err.println("(StackLattice)ERROR! - method difference(FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2) {
		System.err.println("(StackLattice)ERROR! - method difference(FlowSet paramFlowSet1, FlowSet paramFlowSet2)");
		assert(false);
	}
	@Override
	public void intersection(FlowSet paramFlowSet) {
		System.err.println("(StackLattice)ERROR! - method intersection(FlowSet paramFlowSet)");
		assert(false);
	}

	@Override
	public void remove(Object obj, FlowSet flow) {
		System.err.println("(StackLattice)ERROR! - method remove(Object obj, FlowSet flow)");
		assert(false);
	}
	@Override
	public void remove(Object arg) {
		System.err.println("(StackLattice)ERROR! - method remove(Object arg)");
		assert(false);
	}
	public SetIfVars oldVars() {
		 return old_ifVars.copy();
	}
}
