package pt.moth.dependencies.thirdPhase;

import java.util.HashSet;
import java.util.Set;

import pt.moth.dependencies.Dep;
import pt.moth.dependencies.VarDepCollector;
import pt.moth.dependencies.VariableGraph;
import pt.moth.util.View;
import pt.moth.util.graphs.DepEdge;
import pt.moth.util.variables.GlobalVar;
import pt.moth.util.variables.NativeGlobalVar;
import pt.moth.util.variables.ReturnRef;
import pt.moth.util.variables.Variable;
import pt.moth.viewsAnalysis.ResultsCollector;



/**
 * This class computes and stores every dependency between views.A dependency is
 * a triple (V1,var,V2) that correspond to a dependence between views V1 and V2
 * with variable var. This occurs, for example, when we have the following code:
 * var = res(V1); if(...var...){ ...V2... }
 * 
 * This class' "computeViewDeps" method represents the third phase of the
 * dependency analysis
 * 
 * @author Vasco Pessanha
 * 
 */
public class ViewDepCollector {

	private static Set<Dep> dependencies = new HashSet<Dep>();

	public static String printDeps() {
		String r = "\n@@@@@ Dependencies: @@@@@\n";
		for (Dep d : dependencies) {
			r += d.toString() + "\n";
		}
		return r;
	}

	public static Set<Dep> getDependencies() {
		return dependencies;
	}

	/**
	 * For each dependency of the pruned graph (edge) we create a Dep object
	 * that represent the dependency between the 2 views
	 */
	public static void computeViewDeps() {
		VariableGraph graph = VarDepCollector.getPrunnedGraph();
		for(DepEdge edge : graph.edges()){
			Variable origin = edge.getOrigin().getVariable();
			Variable dest = edge.getDest().getVariable();
			View v1 = ResultsCollector.getView(origin.getSootMethod());
			View v2 = ResultsCollector.getView(dest.getSootMethod());
			validPair(origin, v1);
			validPair(dest, v2);
			dependencies.add(new Dep(v1, (GlobalVar) origin, v2, edge.isControlDep()));
		}
	}

	/**
	 * Validates if "var" is a GlobalVariable and "view" is a valid view (not
	 * null) If this conditions is not verified a error message is printed and
	 * the program is stopped
	 */
	private static void validPair(Variable var, View view) {
		boolean isValid = false;
		//NativeGlobalVar
		if(var instanceof NativeGlobalVar){
			isValid = true;
		}
		//is GlobalVar
		if((var instanceof GlobalVar || var instanceof ReturnRef) && view != null){
			isValid = true;
		}
		if (!isValid){
			System.err
					.println("ViewDepCollector - Pair was not Validated! "
							+ "(Shouldn't happen if the graph was well pruned!)\n"
							+ "global = "+ var+ "("	+ var.getClass()+ ")"
							+ "\nview = " + view);
			System.exit(-1);
		}
	}
}
