package pt.moth.sensors;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Map.Entry;


import pt.moth.dependencies.VarDepCollector;
import pt.moth.dependencies.VarVersion;
import pt.moth.dependencies.VariableGraph;
import pt.moth.processAnalysis.Process;
import pt.moth.util.View;
import pt.moth.util.ViewCardinalityComparator;
import pt.moth.util.conflicts.Conflict;
import pt.moth.util.conflicts.ConflictType;
import pt.moth.util.conflicts.ViewConflict;
import pt.moth.util.variables.ParamRef;
import pt.moth.util.variables.ReturnVar;
import pt.moth.util.variables.Variable;
import pt.moth.viewsAnalysis.ViewsCollector;


public class ViewConsistencySensor implements Sensor{
	public final boolean EXCLUDERwRW = false;
	// List of all the processes
	private Set<Process> processes;
	
	// For each Process, these atributes keep the views (Read and Write)
	// of that process
	private Map<String, List<View>> read_views;
	private Map<String, List<View>> write_views;
	
	// For each Process, these atributes keep the maximalViews (Read and Write)
	// of that process
	private Map<String, List<View>> read_maxViews;
	private Map<String, List<View>> write_maxViews;
	private VariableGraph graph;
	
	public ViewConsistencySensor(){
		processes = new HashSet<Process>();
		read_views = new HashMap<String, List<View>>();
		write_views = new HashMap<String, List<View>>();
		read_maxViews = new HashMap<String, List<View>>();
		write_maxViews = new HashMap<String, List<View>>();
	}
	
	@Override
	public void pushViewInformation() {
		//For each process...
		for(Entry<String, List<View>> e : ViewsCollector.getReadViews()){
			read_views.put(e.getKey(), e.getValue());
		}
		for(Entry<String, List<View>> e : ViewsCollector.getWriteViews()){
			write_views.put(e.getKey(), e.getValue());
		}
		for(Process p : ViewsCollector.getProcesses()){
			processes.add(p);
			computeMaximalViews(p.getName());
		}
	}
	
	// Compute Maximal Views...
	private void computeMaximalViews(String process) {
		computeMW(process);
		computeMR(process);
	}
	/**
	 * Compute the Write Maximal Views (MW), i.e., the maximal views of type
	 * Write For each view, see if it is subset of one of the others
	 */
	private void computeMW(String p) {
		boolean isMaximalView;
		List<View> allviews = write_views.get(p);
		List<View> maximalViews = new ArrayList<View>();
		for (View v : allviews) {
			isMaximalView = true;
			for (View v2 : allviews) {
				if (!v2.equals(v) && v2.contains(v)) {
					isMaximalView = false;
					break;
				}
			}
			if (isMaximalView) {
				maximalViews.add(v);
			}
		}
		write_maxViews.put(p, maximalViews);
	}

	/**
	 * Compute the Write Maximal Views (MW), i.e., the maximal views of type
	 * Write For each view, see if it is subset of one of the others
	 */
	private void computeMR(String p) {
		boolean isMaximalView;
		List<View> allviews = read_views.get(p);
		List<View> maximalViews = new ArrayList<View>();
		for (View v : allviews) {
			isMaximalView = true;
			for (View v2 : allviews) {
				if (!v2.equals(v) && v2.contains(v)) {
					isMaximalView = false;
					break;
				}
			}
			if (isMaximalView) {
				maximalViews.add(v);
			}
		}
		read_maxViews.put(p, maximalViews);
	}
	@Override
	public Set<Conflict> getConflicts() {
		boolean allOfTheSameSize = (read_maxViews.size() == write_maxViews
				.size())
				&& (read_maxViews.size() == read_views.size())
				&& (read_maxViews.size() == write_views.size())
				&& (read_maxViews.size() == processes.size());
		Set<Conflict> conflicts = new HashSet<Conflict>();
		if (!allOfTheSameSize) {
			System.err
					.println("Error! (ResultsHandler) Every of this atributes should have the same size (equals to the number of the processes)");
		} else {
			Process[] array = new Process[processes.size()]; // could be with
																// another
																// atribute
			int pos = 0;
			// Store in the array
			for (Process p : processes) {
				array[pos] = p;
				pos++;
			}
			// Match every processor with the other
			for (int i = 0; i < array.length; i++) {
				for (int j = i; j < array.length; j++) {
					getConflicts(conflicts, array[i], array[j]);
					if (i != j) {
						getConflicts(conflicts, array[j], array[i]);
					}
				}
			}
		}
		return conflicts;
	}
	private void getConflicts(Set<Conflict> conflicts, Process p1,
			Process p2) {
		Set<View> overlapping_views;
		List<View> p1_maxViewsW = write_maxViews.get(p1.getName());
		List<View> p1_maxViewsR = read_maxViews.get(p1.getName());
		List<View> p2_ViewsW = write_views.get(p2.getName());
		List<View> p2_ViewsR = read_views.get(p2.getName());

		View intersection = null;
		// Conflict(RwR) Between a Write Maximal View (MW) and Read Views
		if (p1_maxViewsW != null && p2_ViewsR != null) {
			for (View max_v : p1_maxViewsW) {
				overlapping_views = new HashSet<View>();
				for (View v : p2_ViewsR) {
					intersection = v.intersection(max_v);
					if(intersection != null && intersection.size()>0 &&
							!intersection.getMethod().equals(max_v.getMethod())){
						overlapping_views.add(intersection.clone());
					}
				}
				removeCommumVariables(overlapping_views);
				if (!formChain(overlapping_views)) {
					simplify(overlapping_views);
					if(hasCommunDependency(overlapping_views, graph)){
						conflicts.add(new ViewConflict(p1, p2, overlapping_views, max_v,
								ConflictType.RwR));
					}
				}
			}
		}
		// Conflict(WrW) Between a Read Maximal View (MR) and Write Views
		if (p1_maxViewsR != null && p2_ViewsW != null) {
			for (View max_v : p1_maxViewsR) {
				overlapping_views = new HashSet<View>();
				for (View v : p2_ViewsW) {
					intersection = v.intersection(max_v);
					if(intersection != null && intersection.size()>0 &&
							!intersection.getMethod().equals(max_v.getMethod())){
						overlapping_views.add(intersection.clone());
					}
				}
				removeCommumVariables(overlapping_views);
				if (!formChain(overlapping_views)) {
					simplify(overlapping_views);
					conflicts.add(new ViewConflict(p1, p2, overlapping_views, max_v,
						ConflictType.WrW));
				}
			}
		}

	}

	/**
	 * We take a set of views that do not form a chain and simplify it. If a set
	 * of views (v1,v2,..) do not form a chain but, if we take off v1,
	 * (v2,v3,..) still do not form a chain that we can remove v1
	 * 
	 * @param overlapping_views
	 *            - initial set
	 */
	private void simplify(Set<View> overlapping_views) {
		List<View> copy = new ArrayList<View>();
		for(View v : overlapping_views)
			copy.add(v.clone());
		
		for(View v : copy){
			boolean removed = overlapping_views.remove(v);
			if(!removed){
				System.err.println("ViewConsistencySensor, ERROR: Should been removed!");
				System.exit(-1);
			}
			if(formChain(overlapping_views)){
				overlapping_views.add(v);
			}
		}
	}

	private void removeCommumVariables(Set<View> overlapping_views) {
		Set<Variable> comunVars = new HashSet<Variable>();
		Set<Variable> toRemove = null;
		boolean first = true;
		
		for(View v : overlapping_views){
			toRemove = new HashSet<Variable>();
			if(first){
				comunVars = v.getVariables();
				first = false;
			}
			else{
				Set<Variable> vVars = v.getVariables();
				for(Variable var : comunVars){
					if(!vVars.contains(var)){
						toRemove.add(var);
					}
				}
			}
			for(Variable var : toRemove){
				comunVars.remove(var);
			}
			if(comunVars.size() == 0){
				break;
			}
		}
		//Commun Variables can be removed from all the views.
		for(View v : overlapping_views){
			for(Variable var : comunVars){
				v.removeAccessesWithVar(var);	
			}
		}
	}

	/**
	 * Returns if there is a given variable that is reachable from at least two
	 * of the views in "overlapping_views"
	 */
	private boolean hasCommunDependency(Set<View> overlapping_views,
			VariableGraph graph) {
		
		SetOfVars[] reachedVars = new SetOfVars[overlapping_views.size()];
		int count = 0;
		for(View v : overlapping_views){
			ReturnVar returnVar = new ReturnVar(v.getMethod());
			Set<VarVersion> reached = graph.getReachedVars(new VarVersion(returnVar));
			reachedVars[count] = new SetOfVars(reached);
			count++;
		}

		for(int i = 0; i < reachedVars.length; i++){
			SetOfVars v1 = reachedVars[i];
			for(int j = i+1; j < reachedVars.length; j++){
				SetOfVars v2 = reachedVars[j];
				if(hasCommonImportantVertice(v1,v2)){
					return true;
				}
			}
		}
		return false;
	} 
	
	/**
	 * Returns true iff there is a common variable in both sets s1 and s2, and it it an importantVar
	 * @param v1 - First Set
	 * @param v2 - Second Set
	 * @param importantNodes - Set(MAP) with all important nodes
	 */
	private boolean hasCommonImportantVertice(SetOfVars v1, SetOfVars v2) {
		for(VarVersion v : v1){
			if(v2.containsvar(v)){
				return true;
			}
		}
		// If 2 variables go to different parameters of the same method...
		return commonParams(v1.getParams(),v2.getParams());
	}

	private boolean commonParams(Set<ParamRef> set1, Set<ParamRef> set2) {
		for(ParamRef v1 : set1){
			for(ParamRef v2 : set2){
				if(v1.getInvStmt().equals(v2.getInvStmt()) &&
						v1.getSootMethod().equals(v2.getSootMethod())){
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Returns if the given set forms a chain.. For All members mi, the forms a
	 * chain if m0 C m1 C m2 C .. C mn
	 * 
	 * Algorithm: 1) Order the list by view's size. If 2 views have the same
	 * size it's not important which is the bigger 2) Iterate the list comparing
	 * in groups of 2 2.1) If v2.size < v1.size => ERROR! Should be ordered 2.2)
	 * If v2.size >= v1.size => v2.contains(v1)? (Yes => keep going)(No =>
	 * Returns false) 3) Return true at the end
	 * 
	 * 
	 * @param subsetViews
	 *            - the given Set
	 * @return true if the set forms a chain and false otherwise
	 */
	private boolean formChain(Set<View> subsetViews) {
		if (subsetViews.size() <= 1) {
			return true;
		}
		TreeSet<View> orderedSet = new TreeSet<View>(
				new ViewCardinalityComparator());
		for (View v : subsetViews) {
			orderedSet.add(v);
		}

		Object[] array = orderedSet.toArray();
		View v1, v2;
		for (int i = 0; i < orderedSet.size() - 1; i++) {
			v1 = (View) array[i];
			v2 = (View) array[i + 1];
			if (v2.size() < v1.size()) {
				System.err
						.println("ERROR! (Results Handler) - The Set should be ordered...");
				System.exit(-1);
			} else if (v2.size() == v1.size()
					&& !v2.getAccesses().equals(v1.getAccesses())) {
				return false;
			}
			if (!v2.contains(v1)) {
				return false;
			}
		}
		return true;
	}
	// Printing Processes' Maximal Views...
	public String printMaximalViews() {
		String result = "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				+ "@@@  Print Maximal Views  @@@\n"
				+ "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n";

		result += "\nMaximal Read Views (MR)\n";
		for (Entry<String, List<View>> entry : read_maxViews.entrySet()) {
			result += "\n" + entry.getKey() + "'s Views:\n";
			for (View v : entry.getValue()) {
				result += v.toString() + "\n";
			}
		}
		result += "\n\nMaximal Write Views (MW)\n";
		for (Entry<String, List<View>> entry : write_maxViews.entrySet()) {
			result += "\n" + entry.getKey() + "'s Views:\n";
			for (View v : entry.getValue()) {
				result += v.toString() + "\n";
			}
		}
		return result;
	}

	@Override
	public void pushDepInformation() {
		this.graph = VarDepCollector.getGraph();
	}
}
class SetOfVars implements Iterable<VarVersion>{
	protected Set<VarVersion> variables;
	
	public SetOfVars(){
		variables = new HashSet<VarVersion>();
	}
	public Set<ParamRef> getParams() {
		Set<ParamRef> result = new HashSet<ParamRef>();
		for(VarVersion node : variables){
			if(node.getVariable() instanceof ParamRef){
				result.add((ParamRef)node.getVariable().copy());
			}
		}
		return result;
	}
	public SetOfVars(Set<VarVersion> set){
		variables = copy(set);
	}
	
	public boolean containsvar(VarVersion var){
		for(VarVersion v : variables){
			if(v.equals(var)){
				return true;
			}
		}
		return false;
	}
	private Set<VarVersion> copy(Set<VarVersion> set) {
		Set<VarVersion> result = new HashSet<VarVersion>();
		for(VarVersion v : set){
			result.add(v.copy());
		}
		return result;
	}
	
	@Override
	public Iterator<VarVersion> iterator() {
		return variables.iterator();
	}
	
	@Override
	public String toString(){
		if(variables.size() == 0)
			return "{}";
		String r = "{";
		for(VarVersion v : variables){
			r += v + ", ";
		}
		return r.substring(0,r.length()-2)+"}"; 
	}
}