package pt.moth.util.graphs;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;


/**
 * This class represents a directed graph. "vertices" represent the graph's
 * vertices and "edges" represent the graph's edges
 * 
 * @author Vasco Pessanha
 * 
 * @param <E>
 *            Tipe of this graph's vertices
 */
public class DirectedGraph<V,E extends IEdge<V>> {
	protected Set<V> vertices;
	protected Set<E> edges;
	protected Map<V, Set<E>> succs;

	public DirectedGraph() {
		vertices = new HashSet<V>();
		edges = new HashSet<E>();
		succs = new HashMap<V, Set<E>>();
	}

	/**
	 * Adds a new vertice to this graph if it doesn't exist
	 */
	public void addVertex(V v) {
		if (!vertices.contains(v))
			vertices.add(v);
	}

	/**
	 * Adds a new edge to this graph. If the edge's vertices don't exist, they
	 * are inserted.
	 */
	public void addEdge(E e) {
		V origin = e.getOrigin();
		V dest = e.getDest();

		addVertex(origin);
		addVertex(dest);
		edges.add(e);

		addSucc(e);
	}

	private void addSucc(E edge) {
		Set<E> set = succs.get(edge.getOrigin());
		if (set == null) {
			set = new HashSet<E>();
		}
		set.add(edge);
		succs.put(edge.getOrigin(), set);
	}

	private void removeSucc(E e) {
		Set<E> set = succs.get(e.getOrigin());
		if (set != null) {
			set.remove(e);
			succs.put(e.getOrigin(), set);
		}
	}

	/**
	 * Removes the edge "e" from the graph's edges
	 */
	public void removeEdge(E e) {
		edges.remove(e);
		removeSucc(e);
	}

	/**
	 * Removes every edge which origin or dest are equal to the vertice "vertex"
	 */
	public void removeEdgeWithVertex(V vertex) {
		Set<E> toRemove = new HashSet<E>();
		for (E e : edges) {
			if (e.getOrigin().equals(vertex) || e.getDest().equals(vertex)) {
				toRemove.add(e);
			}
		}
		for (E e : toRemove) {
			removeEdge(e);
		}
	}

	/**
	 * Returns a set with all graph's edges
	 */
	public Set<E> edges() {
		return edges;
	}

	/**
	 * Returns true iff this graph contains the edge "e" and false otherwise
	 */
	public boolean containsEdge(E e) {
		return edges.contains(e);
	}

	/**
	 * Returns if this graph has no edges
	 */
	public boolean isEmpty() {
		return edges.isEmpty();
	}

	/**
	 * Returns the number of edges of this graph
	 */
	public int size() {
		return edges.size();
	}

	/**
	 * Resets this graph
	 */
	public void clear() {
		vertices = new HashSet<V>();
		edges = new HashSet<E>();
	}

	public String toString() {
		String r = "";
		for (E e : edges) {
			r += e.toString() + "\n";
		}
		return r;
	}

	public void printGraph() {
		for (E e : edges) {
			System.out.println(e.toString());
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		for (E e : edges) {
			result = prime * result + ((edges == null) ? 0 : e.hashCode());
		}
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof DirectedGraph))
			return false;
		@SuppressWarnings("unchecked")
		DirectedGraph<V,E> other = (DirectedGraph<V,E>) obj;
		return sameEdges(other);
	}

	/**
	 * Returns true iff this graph has exactly the same edges as the graph
	 * "other". Otherwise, it returns false.
	 */
	private boolean sameEdges(DirectedGraph<V,E> other) {
		if (size() != other.size())
			return false;

		for (E v : other.edges) {
			if (!edges.contains(v)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Return this graph's edges' iterator
	 */
	public Iterator<E> iterator() {
		return edges.iterator();
	}

	/**
	 * Return all the vertex's direct successors
	 */
	public Set<E> getSuccEdges(V vertex) {
		Set<E> succEdges = succs.get(vertex);
		if (succEdges != null){
			return succEdges;
		}
		return new HashSet<E>();
	}


	/**
	 * Return all the predecessors of "v". Complexity = O(E), E - number of
	 * edges
	 */
	public Set<E> getPredEdges(V v) {
		Set<E> result = new HashSet<E>();
		for (E e : edges) {
			if (e.getDest().equals(v)) {
				result.add(e);
			}
		}
		return result;
	}
	
	/**
	 * Return all the vertex's direct successors
	 */
	public Set<V> getSuccs(V vertex) {
		Set<E> succEdges = succs.get(vertex);
		Set<V> result = new HashSet<V>();
		if (succEdges != null){
			for(E e : succEdges){
				result.add(e.getDest());
			}
		}
		return result;
	}


	/**
	 * Return all the predecessors of "v". Complexity = O(E), E - number of
	 * edges
	 */
	public Set<V> getPreds(V v) {
		Set<V> result = new HashSet<V>();
		for (E e : edges) {
			if (e.getDest().equals(v)) {
				result.add(e.getOrigin());
			}
		}
		return result;
	}
	
	/**
	 * Returns true iff this graph contains the vertex "v"
	 */
	public boolean containsVertex(V v) {
		return vertices.contains(v);
	}

	/**
	 * Return a set with all graph vertices
	 */
	public Set<V> getVertices() {
		return vertices;
	}

	/**
	 * Return a set of all vertices that can be reached from "v" Reached(v1,v2)
	 * :- Edge(v1,v2) || (Edge(v1,v') AND Reached(v',v2)
	 */
	public Set<V> getReachedVertices(V v) {
		Set<V> result = new HashSet<V>();
		for (V vertex : vertices) {
			if (this.hasPath(v, vertex)) {
				result.add(vertex);
			}
		}
		return result;
	}

	/**
	 * Returns whether or not there is a path from the origin vertice to dest
	 * 
	 * @param origin
	 *            - first vertice
	 * @param dest
	 *            - second vertice
	 * @return true (there is a path) or false (there is not a path)
	 */
	public boolean hasPath(V origin, V dest) {
		Set<V> treated = new HashSet<V>();
		List<V> toExplore = new ArrayList<V>();
		toExplore.add(origin);
		while (toExplore.size() > 0) {
			V v = toExplore.remove(0);
			if (!treated.contains(v)) {
				treated.add(v);
				Set<V> succs = getSuccs(v);
				for (V e : succs) {
					if (e.equals(dest)) {
						return true;
					} else if (!treated.contains(e)) {
						toExplore.add(e);
					}
				}
			}
		}
		return false;
	}
	
	/**
	 * Removes the vertice from the "vertices" attribute
	 */
	public void removeVertex(V vertex) {
		vertices.remove(vertex);
	}
	
}
