package pt.moth.util.variables;

import soot.SootMethod;
import soot.jimple.FieldRef;
import soot.jimple.StaticFieldRef;
import soot.jimple.internal.JInstanceFieldRef;

/**
 * This Class represents a global variable. Despite the fact that the methods
 * equals and hashCode are implemented using only the name atribute, we keep the
 * Soot Objects for future using of their information
 * 
 * The atributes are:
 * 
 * @name, is the name of this variable, "class.name" format, generated with the
 *        other two atributes
 * @jinstance, Soot Object that has a more information about this variable
 * 
 * @author Vasco Pessanha
 * 
 */
public class GlobalVar implements IGlobalVar{

	// Name of this variable
	protected String name;
	// Soot object for this variable
	protected FieldRef jinstance;
	// Method (optional) where this variable was accessed.
	protected SootMethod method;
	// Determines if this variable was accessed directly in "method", or was
	// accessed in a non-atomic method called by "method"
	protected boolean direct;

	// Note:  The last two attributes ("method" and "direct") are only used in the dependency analysis
	public GlobalVar(GlobalVar v) {
		this.jinstance = v.getJinstance();
		this.method = v.method;
		this.direct = v.direct;
		this.name = v.name;
	}

	public GlobalVar(FieldRef jinstance) {
		this(jinstance,null,true);
	}
	public GlobalVar(FieldRef jinstance,boolean isDirect) {
		this(jinstance,null,isDirect);
	}
	public GlobalVar(FieldRef jinstance, SootMethod m) {
		this(jinstance,m,true);
	}
	public GlobalVar(FieldRef jinstance, SootMethod m, boolean isDirect) {
		if(jinstance != null && !(jinstance instanceof JInstanceFieldRef) && !(jinstance instanceof StaticFieldRef)){
			System.out.println("(GlobalVar) ERROR! - This shouldn't happen! class = "+jinstance.getClass());
			System.exit(-1);
		}
		this.jinstance = jinstance;
		this.method = m;
		this.direct = isDirect;
		this.name = generateName();
	}
	private String generatesSimplifiedName() {
		if (jinstance == null){
			return "";
		}
		else if(jinstance.getField() == null){
			System.err.println("GlobalVar : ERROR! - Shouldn't happen ("+jinstance+")");
			return "";
		}
		String prefix = jinstance.getField().getDeclaringClass().getName();
		String sufix = 	jinstance.getField().getName();
		return prefix+ "." +sufix;
		
	}
	private String generateName() {
		if (jinstance == null){
			return "";
		}
		else if(jinstance.getField() == null){
			System.err.println("GlobalVar : ERROR! - Shouldn't happen ("+jinstance+")");
			return "";
		}
		String prefix = jinstance.getField().getDeclaringClass().getName();
		String sufix = 	jinstance.getField().getName();
		if(method == null){
			return prefix+ "." +sufix+ (direct?"":"(I)");
		}
		return prefix+ "." +sufix+ "(" + method.getName()+(direct?"":" - I")+")";
		
	}
	public String toString(){
		return name;
	}
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public FieldRef getJinstance() {
		return jinstance;
	}

	public void setJinstance(JInstanceFieldRef jinstance) {
		this.jinstance = jinstance;
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((method == null) ? 0 : method.hashCode());
		result = prime * result + ((name == null) ? 0 : name.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (!(obj instanceof GlobalVar))
			return false;
		GlobalVar other = (GlobalVar) obj;
		if (method == null) {
			if (other.method != null)
				return false;
		} else if (!method.equals(other.method))
			return false;
		if (name == null) {
			if (other.name != null)
				return false;
		} else if (!name.equals(other.name))
			return false;
		return true;
	}

	public GlobalVar copy(){
		return new GlobalVar(jinstance,method,direct);
	}
	
	public boolean hasSootMethod(){
		return method != null;
	}

	@Override
	public SootMethod getSootMethod() {
		return method;
	}

	public boolean equalsIgnoreMethod(GlobalVar v) {
		if(jinstance == null){
			return v.getFieldRef() == null;
		}
		return generatesSimplifiedName().equals(v.generatesSimplifiedName());
	}

	public FieldRef getFieldRef() {
		return jinstance;
	}

	public boolean isDirect() {
		return direct;
	}
}
