package pt.moth.util.variables;

import soot.SootMethod;
import soot.jimple.Stmt;

/**
 * This class represents a Parameter Reference of a given method call.
 * This type of variables has no sucessors, and its unique predecessor is the variable with with the method was called
 * However, when we are prunning we jump from every ParamRef? to the real ParamVar?
 * 
 * @author Vasco Pessanha
 */
public class ParamRef implements Variable{

	protected int num;
	protected SootMethod callerMethod;
	protected SootMethod calledMethod;
	protected Stmt invStmt;
	protected String name;
	
	public ParamRef(SootMethod callerMethod, int num, Stmt invStmt){
		assert(num >= 0 && invStmt.containsInvokeExpr());
		
		this.callerMethod = callerMethod;
		this.calledMethod = invStmt.getInvokeExpr().getMethod();
		this.invStmt = invStmt;
		this.num = num;
		this.name = generateName();
	}

	private String generateName() {
		return calledMethod.getName()+".ParamRef"+num;
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public Variable copy() {
		return new ParamRef(callerMethod, num, invStmt);
	}

	@Override
	public SootMethod getSootMethod() {
		return calledMethod;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((callerMethod == null) ? 0 : callerMethod.hashCode());
		result = prime * result + ((invStmt == null) ? 0 : invStmt.hashCode());
		result = prime * result + num;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ParamRef other = (ParamRef) obj;
		if (callerMethod == null) {
			if (other.callerMethod != null)
				return false;
		} else if (!callerMethod.equals(other.callerMethod))
			return false;
		if (invStmt == null) {
			if (other.invStmt != null)
				return false;
		} else if (!invStmt.equals(other.invStmt))
			return false;
		if (num != other.num)
			return false;
		return true;
	}
	
	
	/**
	 * Returns the caller method (the method where the method call was performed) 
	 */
	public SootMethod getCallerMethod() {
		return callerMethod;
	}

	/**
	 * Returns the invoke statement of this method call
	 */
	public Stmt getInvStmt() {
		return invStmt;
	}

	/**
	 * Returns the actual ParamVar of that method
	 */
	public ParamVar getParamVar() {
		 return new ParamVar(calledMethod, num);
	}
	
	public String toString(){
		return name;
	}
}
