package pt.moth.util.variables;

import soot.SootMethod;
import soot.jimple.Stmt;

/**
 * This class represents a Return reference of a given method. Therefore, we can
 *  have a graph of all method's dependencies and then, when we have
 *  "x = method()", we add the dependency "method.ReturnRef => x". With this, we
 *  get fewer false positives due to dependencies between every left variable of
 *  a method call and its ReturnVar
 * When we are prunning we jump from every ReturnVar to its specific ReturnRef or
 *  to all ReturnRef's otherwise
 * 
 * @author Vasco Pessanha
 */
public class ReturnRef implements Variable {

	protected SootMethod calledMethod;
	protected SootMethod callerMethod;
	protected Stmt invStmt;
	protected String name;

	public ReturnRef(SootMethod callerMethod, Stmt stmt) {
		assert (stmt.containsInvokeExpr());
		this.calledMethod = stmt.getInvokeExpr().getMethod();
		this.callerMethod = callerMethod;
		this.invStmt = stmt;
		this.name = generateName();
	}

	private String generateName() {
		return "ReturnRef(" + callerMethod.getName() + ","
				+ calledMethod.getName() + ")";
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public ReturnRef copy() {
		return new ReturnRef(callerMethod, invStmt);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((invStmt == null) ? 0 : invStmt.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ReturnRef other = (ReturnRef) obj;
		if (invStmt == null) {
			if (other.invStmt != null)
				return false;
		} else if (!invStmt.equals(other.invStmt))
			return false;
		return true;
	}

	@Override
	public SootMethod getSootMethod() {
		return calledMethod;
	}

	public String toString() {
		return name;
	}
}
