package pt.moth.viewsAnalysis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import pt.moth.processAnalysis.Process;
import pt.moth.util.Access;
import pt.moth.util.IAccess;
import pt.moth.util.View;
import pt.moth.util.variables.Variable;
import soot.SootMethod;

/**
 * Class that collects all processes' read and write maximal views.
 * This information will be provided to all the implemented sensors
 * 
 * @author Vasco Pessanha
 *
 */
public class ViewsCollector {

	// List of all the processes
	private static Set<Process> processes = new HashSet<Process>();
	// For each Process this attribute keeps its read views
	private static Map<String, List<View>> read_views = new HashMap<String, List<View>>();
	// For each Process this attribute keeps its write views
	private static Map<String, List<View>> write_views = new HashMap<String, List<View>>();
	
	//variables to exclude/not consider in the view creation
	private static final String[] VARIABLES_TO_EXCLUDE = {
			"java.util.concurrent.atomic.AtomicLong",
			"java.lang.Throwable"};
		
	/**
	 * Add Process (with its Views)...
	 * @param p process being added
	 */
	public static void putProcess(Process p) {
		if(!ResultsCollector.hasResults()){
			System.err.println("ViewsCollector - ERROR! Try to put process before compute views analysis");
			return;
		}
		
		List<View> read_list = new ArrayList<View>();
		List<View> write_list = new ArrayList<View>();
		
		Set<SootMethod> methods = p.getMethods();
		for(SootMethod m : methods){
			Lattice lat = ResultsCollector.getLattice(m.getSignature());
			View rv = extractReadView(m, lat);
			View rw = extractWriteView(m, lat);
			if(rv != null){
				read_list.add(rv);
			}
			if(rw != null){
				write_list.add(rw);
			}
		}
		
		
		if (read_list.size() + write_list.size() > 0) {
			processes.add(new Process(p));
			read_views.put(p.getName(), read_list);
			write_views.put(p.getName(), write_list);
		}
	}
	private static View extractReadView(SootMethod method, Lattice lat) {
		Lattice read_lattice = new Lattice();
		for (IAccess a : lat.toList()) {
			if (a instanceof Access && ((Access) a).isRead() && !toExclude(a.getVariable())) {
				read_lattice.add(a.copy());
			}
		}
		if (read_lattice.size() <= 0) {
			return null;
		}
		return new View(read_lattice, method);
	}
	private static View extractWriteView(SootMethod method, Lattice lat) {
		Lattice write_lattice = new Lattice();
		for (IAccess a : lat.toList()) {
			if (a instanceof Access && ((Access)a).isWrite() && !toExclude(a.getVariable())) {
				write_lattice.add(a.copy());
			}
		}
		if (write_lattice.size() <= 0) {
			return null;
		}
		return new View(write_lattice, method);
	}

	/**
	 * Determines whether or not a given variable should be excluded
	 * @return true - should be excluded, false - should not be excluded
	 */
	private static boolean toExclude(Variable variable) {
		for(String s : VARIABLES_TO_EXCLUDE){
			if(variable == null || variable.toString().contains(s)){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Return all read views needed by each sensor
	 * @return set of entries of type (Process -> Views)
	 */
	public static Set<Entry<String, List<View>>> getReadViews(){
		return read_views.entrySet();
	}
	/**
	 * Return all write views needed by each sensor
	 * @return set of entries of type (Process -> Views)
	 */
	public static Set<Entry<String, List<View>>> getWriteViews(){
		return write_views.entrySet();
	}
	/**
	 * Return all processes
	 * @return all computed processes
	 */
	public static Set<Process> getProcesses(){
		return processes;
	}
	// Printing all methodes' views
	public static String printMethodsViews() {
		String result = "@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"
				+ "@@  Print Methods Views  @@\n"
				+ "@@@@@@@@@@@@@@@@@@@@@@@@@@@\n\nRead Views:\n";

		for (Entry<String, List<View>> e : read_views.entrySet()) {
			result +=e.getKey()+" => "+e.getValue() + "\n";
		}
		result +="\n\nWrite Views:\n";
		for (Entry<String, List<View>> e : write_views.entrySet()) {
			result +=e.getKey()+" => "+e.getValue() + "\n";
		}
		return result;
	}
}
